@@ -15,6 +15,8 @@
  * along with this program.  If not, see
  * `http://www.gnu.org/licenses/'.
  */
+#include "config.h"
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <glib.h>
@@ -27,7 +29,6 @@
 #include <unistd.h>
 #include <fcntl.h>
 #include <errno.h>
-#include "config.h"
 #include <assert.h>
 #include <libgen.h>
 
@@ -61,9 +62,13 @@ static GtkListStore *list_store;
 static char *init_filter;
 static gboolean show_version;
 static gboolean multiselect;
+static gboolean align_left;
+static gboolean print_pattern;
 static char *project_type;
 static char *project_dir;
+static char *pattern_text;
 static gboolean read_stdin;
+static gboolean output_index;
 static gboolean disable_bzr;
 static gboolean disable_hg;
 
@@ -195,8 +200,8 @@ void setup_column(void)
 	GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
 	GtkTreeViewColumn *col;
 	g_object_set(G_OBJECT(renderer),
-		     "ellipsize", PANGO_ELLIPSIZE_START,
-		     "alignment", PANGO_ALIGN_RIGHT,
+		     "ellipsize", align_left ? PANGO_ELLIPSIZE_END : PANGO_ELLIPSIZE_START,
+		     "alignment", align_left ? PANGO_ALIGN_LEFT : PANGO_ALIGN_RIGHT,
 		     NULL);
 
 	col = gtk_tree_view_column_new_with_attributes("filename", renderer, NULL);
@@ -234,7 +239,11 @@ void selection_printer(gpointer data,
 	gtk_tree_model_get_iter(GTK_TREE_MODEL(list_store), &iter, path);
 	gtk_tree_model_get(GTK_TREE_MODEL(list_store), &iter, 0, &idx, -1);
 
-	fputs(files[idx].p, stdout);
+	if (output_index) {
+		printf("%d", idx);
+	} else {
+		fputs(files[idx].p, stdout);
+	}
 
 	gtk_tree_path_free(path);
 }
@@ -246,8 +255,11 @@ void print_selection(void)
 	GList *list;
 
 	list = gtk_tree_selection_get_selected_rows(sel, 0);
-	if (!list)
+	if (!list) {
+		if (print_pattern)
+			fputs(pattern_text, stdout);
 		return;
+	}
 
 	g_list_foreach(list, selection_printer, NULL);
 	g_list_free(list);
@@ -281,9 +293,9 @@ void on_entry_changed(GtkEditable *editable,
 
 	timing_t start = start_timing();
 
-	char *text = g_strdup(gtk_entry_get_text(GTK_ENTRY(editable)));
-	filter_tree_view(text);
-	free(text);
+	free(pattern_text);
+	pattern_text = g_strdup(gtk_entry_get_text(GTK_ENTRY(editable)));
+	filter_tree_view(pattern_text);
 
 	finish_timing(start, "filtration");
 }
@@ -328,7 +340,7 @@ int my_popen(char *string, GPid *child_pid)
 	gboolean ok = g_spawn_async_with_pipes(0, // work dir
 					       argv,
 					       0, //envp
-					       0, // flags
+					       G_SPAWN_DO_NOT_REAP_CHILD, // flags
 					       0, 0, //child setup & user data
 					       child_pid,
 					       0, // stdin
@@ -346,8 +358,9 @@ gpointer setup_filenames_core(gpointer _dummy,
 {
 	int pipe = -1;
 
+	do_with_main_loop_init_complete(p);
+
 	if (project_type && !strcmp(project_type, "mlocate")) {
-		do_with_main_loop_init_complete(p);
 		int fd = open(project_dir, O_RDONLY);
 		if (fd < 0) {
 			perror("setup_filenames:open");
@@ -365,6 +378,8 @@ gpointer setup_filenames_core(gpointer _dummy,
 		if (!find_command)
 			find_command = FIND_INVOCATION;
 		pipe = my_popen(find_command, &child_pid);
+	} else if (!strcmp(project_type, "script")) {
+		pipe = my_popen("./.gpicker-script", &child_pid);
 	} else if (!strcmp(project_type, "git"))
 		pipe = my_popen("git ls-files --exclude-standard -c -o -z .", &child_pid);
 	else if (!strcmp(project_type, "hg"))
@@ -372,7 +387,6 @@ gpointer setup_filenames_core(gpointer _dummy,
 	else if (!strcmp(project_type, "bzr"))
 		pipe = my_popen("bzr ls -R --versioned --unknown --null", &child_pid);
 
-	do_with_main_loop_init_complete(p);
 
 	if (pipe < 0) {
 		perror("failed to spawn find");
@@ -381,12 +395,27 @@ gpointer setup_filenames_core(gpointer _dummy,
 
 	read_filenames(pipe);
 
-	if (child_pid) {
-		kill(child_pid, SIGINT);
-		child_pid = 0;
-	}
 	close(pipe);
 
+	if (child_pid) {
+		pid_t rv;
+		int child_status;
+		do {
+			rv = waitpid(child_pid, &child_status, 0);
+			if (rv < 0) {
+				if (errno == EINTR)
+					continue;
+				perror("waitpid");
+				exit(1);
+			}
+		} while (0);
+		child_pid = 0;
+		if (WIFEXITED(child_status) && (child_status = WEXITSTATUS(child_status)) != 0) {
+			fprintf(stderr, "child exited with bad status: %d\n", child_status);
+			exit(1);
+		}
+	}
+
 	return 0;
 }
 
@@ -476,8 +505,13 @@ GOptionEntry entries[] = {
 	{"dir-separator", 'd', 0, G_OPTION_ARG_STRING, &dir_separator, "separator of directory names from stdin (/ is default)", 0},
 	{"eat-prefix", 0, 0, G_OPTION_ARG_STRING, &eat_prefix, "eat this prefix from names (./ is default)", 0},
 	{"multiselect", 'm', 0, G_OPTION_ARG_NONE, &multiselect, "enable multiselect", 0},
+	{"left-align", 'l', 0, G_OPTION_ARG_NONE, &align_left, "left align everything (default is right-align)", 0},
+	{"print-pattern", 'p', 0, G_OPTION_ARG_NONE, &print_pattern, "print pattern if nothing matched (default is false)", 0},
+	{"dont-sort", 'S', 0, G_OPTION_ARG_NONE, &dont_sort_initial, "dont sort result list", 0},
 	{"init-filter", 'i', 0, G_OPTION_ARG_STRING, &init_filter, "initial filter value", 0},
 	{"load-stdin-too", 0, 0, G_OPTION_ARG_NONE, &gpicker_load_stdin_too, "read additional filenames from stdin", 0},
+	{"output-index", 'I', 0, G_OPTION_ARG_NONE, &output_index, "print selection index instead of value (implies -S)", 0},
+	{"ignore-positions", 'P', 0, G_OPTION_ARG_NONE, &ignore_positions, "ignore match position for sorting", 0},
 	{0}
 };
 
@@ -521,6 +555,17 @@ int check_parents(char* name)
 	return rv;
 }
 
+static
+int detect_project_type_script(void)
+{
+	struct stat st;
+	int rv = stat(".gpicker-script", &st);
+	/* NOTE: I was considering checking if script is executable
+	 * but decided that guess semantics shouldn't bother with
+	 * it. It provides better error reporting. */
+	return (rv == 0 && S_ISREG(st.st_mode));
+}
+
 static
 void enter_project_dir()
 {
@@ -532,7 +577,9 @@ void enter_project_dir()
 	}
 
 	if (project_type && !strcmp(project_type, "guess")) {
-		if (check_parents(".git"))
+		if (detect_project_type_script())
+			project_type = "script";
+		else if (check_parents(".git"))
 			project_type = "git";
 		else if (!disable_hg && check_parents(".hg"))
 			project_type = "hg";
@@ -583,6 +630,10 @@ void parse_options(int argc, char **argv)
 		fprintf(stderr, "option parsing failed: %s\n", error->message);
 		exit(1);
 	}
+
+	if (output_index)
+		dont_sort_initial = 1;
+
 	if (show_version) {
 		fprintf(stdout, "%s %s\n", PACKAGE_NAME, VERSION);
 		exit(0);
@@ -593,11 +644,12 @@ void parse_options(int argc, char **argv)
 		exit(1);
 	}
 	if (project_type && strcmp(project_type, "guess") &&
-			strcmp(project_type, "git") &&
-			strcmp(project_type, "hg") &&
-			strcmp(project_type, "bzr") &&
-			strcmp(project_type, "default") &&
-			strcmp(project_type, "mlocate")) {
+	    strcmp(project_type, "git") &&
+	    strcmp(project_type, "hg") &&
+	    strcmp(project_type, "bzr") &&
+	    strcmp(project_type, "default") &&
+	    strcmp(project_type, "script") &&
+	    strcmp(project_type, "mlocate")) {
 		fprintf(stderr, "Unknown project type specified: %s\n", project_type);
 		exit(1);
 	}
@@ -648,7 +700,7 @@ void build_ui()
 	gtk_container_add(GTK_CONTAINER(top_window), GTK_WIDGET(vbox));
 
 	name_entry = GTK_ENTRY(gtk_entry_new());
-	gtk_entry_set_alignment(name_entry, 1.0);
+	gtk_entry_set_alignment(name_entry, align_left ? 0.0 : 1.0);
 	gtk_box_pack_start(GTK_BOX(vbox), GTK_WIDGET(name_entry),
 			   FALSE, TRUE, 0);
 
@@ -745,5 +797,8 @@ int main(int argc, char **argv)
 
 	if (!program_exited)
 		gtk_main();
+
+	uninit_filter();
+
 	return 0;
 }
